#!/usr/bin/env node
"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const commandLineArgs = require("command-line-args");
const commandLineUsage = require("command-line-usage");
const fs_1 = require("fs");
const path_1 = require("path");
const api_1 = require("./api");
const ProjectInputOutput_1 = require("./rto-factory/ProjectInputOutput");
process.on("uncaughtException", err => {
    console.error("uncaughtException", err);
    process.exit(1);
});
process.on("unhandledRejection", err => {
    console.trace("unhandledRejection", err);
    process.exit(1);
});
class InvalidArgumentError extends Error {
    constructor(message) {
        super(message);
        this.causeCode = "invalidArgument";
    }
}
const optionDefinitions = [
    {
        name: "help",
        alias: "h",
        type: Boolean,
        description: "Print this help message.",
    },
    {
        name: "output-dir",
        alias: "o",
        type: String,
        description: "The output directory (optional).",
        typeLabel: "{underline directory}"
    },
    {
        name: "source-dir",
        alias: "s",
        type: String,
        description: "The source directory (optional when used with option {underline --ast} or with a single source file).",
        typeLabel: "{underline directory}"
    },
    {
        name: "encoding",
        alias: "e",
        type: String,
        description: "Encoding for input and output file(s) (default is {underline utf8})."
    },
    {
        name: "bundle",
        alias: "b",
        type: String,
        description: "Generate a bundle file for RTO data (optional)."
    },
    {
        name: "prettify",
        type: Boolean,
        description: "Prettify RTO files (optional)."
    },
    {
        name: "ast",
        type: Boolean,
        description: "Generate AST files instead of RTO files (optional)."
    },
    {
        name: "src",
        description: "Input files to process (by default at last position).",
        type: String,
        multiple: true,
        defaultOption: true,
        typeLabel: "{underline file} ..."
    }
];
cli().catch(error => {
    console.error(`Error: ${error.message}`, error);
});
async function cli() {
    const options = parseOptions();
    if (!options)
        return;
    if (options["help"]) {
        printHelp();
        return;
    }
    try {
        await processFiles(options);
    }
    catch (error) {
        if (error.causeCode === "invalidArgument") {
            console.error(`Error: ${error.message}`);
            printHelp();
        }
        else
            throw error;
    }
}
function printHelp() {
    const sections = [
        {
            header: "TypeOnly",
            content: "TypeOnly Parser."
        },
        {
            header: "Synopsis",
            content: [
                "$ npx typeonly {bold --source-dir} {underline src/} {underline file-name.d.ts}",
                "$ npx typeonly {bold --help}"
            ]
        },
        {
            header: "Options",
            optionList: optionDefinitions
        },
        {
            content: "Project home: {underline https://github.com/tomko-team/typeonly}"
        }
    ];
    const usage = commandLineUsage(sections);
    console.log(usage);
}
function parseOptions() {
    try {
        return commandLineArgs(optionDefinitions);
    }
    catch (error) {
        console.log(`Error: ${error.message}`);
        printHelp();
    }
}
async function processFiles(options) {
    if (options["ast"]) {
        if (!options["src"])
            throw new InvalidArgumentError("Missing source file(s).");
        options["src"].map((file) => createAstJsonFile(file, options));
    }
    else
        await createRtoJsonFiles(options);
}
function createAstJsonFile(file, options) {
    file = options["source-dir"] ? path_1.join(options["source-dir"], file) : file;
    const encoding = options["encoding"] || "utf8";
    let source;
    try {
        source = fs_1.readFileSync(file, { encoding });
    }
    catch (err) {
        throw new InvalidArgumentError(`Cannot read file: ${file}`);
    }
    let fileName = path_1.basename(file);
    if (fileName.endsWith(".ts"))
        fileName = fileName.substring(0, fileName.length - (fileName.endsWith(".d.ts") ? 5 : 3));
    const ast = api_1.parseTypeOnly({ source });
    const outFile = path_1.join(options["output-dir"] || path_1.dirname(file), `${fileName}.ast.json`);
    fs_1.writeFileSync(outFile, JSON.stringify(ast, undefined, "\t"), encoding);
}
async function createRtoJsonFiles(options) {
    var _a, _b, _c;
    let srcList = (_a = options["src"]) !== null && _a !== void 0 ? _a : [];
    let sourceDir;
    if (!options["source-dir"]) {
        if (srcList.length === 1)
            sourceDir = path_1.dirname(srcList[0]);
        else
            throw new Error("Missing 'source-dir' option.");
    }
    else
        sourceDir = options["source-dir"];
    sourceDir = normalizeDir(sourceDir);
    if (srcList.length === 0)
        srcList = getTypingFilesInDir(sourceDir);
    const modulePaths = normalizeModulePaths(srcList, sourceDir);
    const encoding = (_b = options["encoding"]) !== null && _b !== void 0 ? _b : "utf8";
    const prettify = options["prettify"] ? "\t" : undefined;
    let bundleName = options["bundle"];
    if (bundleName) {
        if (!bundleName.endsWith(".to.json"))
            bundleName += ".to.json";
        let outputDir = options["output-dir"] ? normalizeDir(options["output-dir"]) : undefined;
        const parsed = path_1.parse(bundleName);
        if (parsed.dir) {
            outputDir = outputDir ? path_1.join(outputDir, parsed.dir) : parsed.dir;
            bundleName = parsed.base;
        }
        if (outputDir) {
            await ProjectInputOutput_1.ensureDirectory(outputDir);
            bundleName = path_1.join(outputDir, bundleName);
        }
        const rtoModules = await api_1.generateRtoModules({
            modulePaths,
            readFiles: {
                sourceDir,
                encoding,
            },
            returnRtoModules: true
        });
        fs_1.writeFileSync(bundleName, JSON.stringify(rtoModules, undefined, prettify), { encoding });
    }
    else {
        const outputDir = normalizeDir((_c = options["output-dir"]) !== null && _c !== void 0 ? _c : sourceDir);
        await api_1.generateRtoModules({
            modulePaths,
            readFiles: {
                sourceDir,
                encoding,
            },
            writeFiles: {
                outputDir,
                prettify
            }
        });
    }
}
function getTypingFilesInDir(dir) {
    const files = fs_1.readdirSync(dir);
    return files
        .filter(fileName => fileName.endsWith(".d.ts"));
}
function normalizeModulePaths(files, sourceDir) {
    const prefix = `${sourceDir}/`.replace(/\\/g, "/");
    return files.map(file => {
        file = file.replace(/\\/g, "/");
        if (file.startsWith(prefix))
            file = `./${file.substr(prefix.length)}`;
        else if (!file.startsWith("./") && !file.startsWith("../"))
            file = `./${file}`;
        return file;
    });
}
function normalizeDir(path) {
    return path.replace(/\\/g, "/").replace(/\/+$/, "");
}
