"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
class CommentGrabber {
    constructor(parsingContext) {
        this.parsingContext = parsingContext;
        this.consumed = new Set();
        this.lastTokenIndex = -1;
    }
    grabCommentsOf(ctx) {
        const { tokenStream, tokenTypes: { NEWLINE } } = this.parsingContext;
        this.lastTokenIndex = ctx.stop.tokenIndex;
        const tokens = getHiddenTokens({ tokenStream, tokenIndex: ctx.start.tokenIndex, direction: "left", NEWLINE });
        const { docComment, standaloneComments } = this.grabDocAndStandaloneComments(tokens);
        const inlineComments = this.grabInlineComments(ctx);
        const result = {
            standaloneCommentsBefore: standaloneComments,
            inlineComments,
        };
        if (docComment)
            result.docComment = docComment.text;
        return result;
    }
    grabStandaloneCommentsAfterLast(ctx) {
        const { tokenStream, tokenTypes: { NEWLINE } } = this.parsingContext;
        const tokens = getHiddenTokens({
            tokenStream,
            tokenIndex: this.lastTokenIndex,
            direction: "right",
            NEWLINE,
            startIndex: ctx ? ctx.start.start : undefined,
            stopIndex: ctx ? ctx.stop.stop : undefined,
        });
        const { standaloneComments } = this.grabDocAndStandaloneComments(tokens, "standaloneOnly");
        return standaloneComments;
    }
    grabDocAndStandaloneComments(tokens, mode) {
        const blocks = this.tokensToBlocks(tokens);
        for (const block of blocks) {
            if (block.syntax === "inline")
                this.consumeTokens(block.singleLineComments);
            else
                this.consumeTokens([block.multiLineComment]);
        }
        return this.blocksToDocAndStandaloneComments(blocks, mode);
    }
    tokensToBlocks(tokens) {
        if (!tokens)
            return [];
        const { tokenTypes: { MULTILINE_COMMENT: ML_COM, SINGLE_LINE_COMMENT: SL_COM, NEWLINE } } = this.parsingContext;
        tokens = keepWholeLineComments(tokens, this.parsingContext);
        const blocks = [];
        let curSingleLinesBlock;
        for (const token of tokens) {
            if (this.consumed.has(token))
                continue;
            if (token.type === ML_COM) {
                blocks.push({
                    syntax: "classic",
                    multiLineComment: token
                });
                curSingleLinesBlock = undefined;
            }
            else if (token.type === NEWLINE) {
                curSingleLinesBlock = undefined;
            }
            else if (token.type === SL_COM) {
                if (!curSingleLinesBlock) {
                    curSingleLinesBlock = {
                        syntax: "inline",
                        singleLineComments: []
                    };
                    blocks.push(curSingleLinesBlock);
                }
                curSingleLinesBlock.singleLineComments.push(token);
            }
        }
        return blocks;
    }
    blocksToDocAndStandaloneComments(blocks, mode) {
        const standaloneComments = [];
        let docComment;
        for (const block of blocks) {
            if (block.syntax === "inline") {
                const comment = this.singleLineTokensToStandaloneComment(block.singleLineComments);
                if (comment)
                    standaloneComments.push(comment);
            }
            else {
                const docOrSa = this.multiLineTokenToDocOrStandaloneComment(block.multiLineComment);
                if (docOrSa) {
                    const { doc, comment } = docOrSa;
                    if (doc && mode !== "standaloneOnly")
                        docComment = comment;
                    else
                        standaloneComments.push(comment);
                }
            }
        }
        return {
            standaloneComments,
            docComment
        };
    }
    singleLineTokensToStandaloneComment(tokens) {
        if (tokens.length === 0)
            return;
        const { source } = this.parsingContext;
        const text = tokens
            .map(({ start, stop }) => formatInlineComment(source, start, stop, "asBlockLine"))
            .join("\n");
        if (text)
            return { syntax: "inline", text };
    }
    multiLineTokenToDocOrStandaloneComment(token) {
        const { source } = this.parsingContext;
        const { start, stop } = token;
        const { doc, text } = formatMultiLineComment(source, start, stop);
        if (text)
            return { doc, comment: { syntax: "classic", text } };
    }
    grabInlineComments(ctx) {
        const { tokenStream, tokenTypes: { COMMA, SEMICOLON, MULTILINE_COMMENT: ML_COM, SINGLE_LINE_COMMENT: SL_COM, NEWLINE } } = this.parsingContext;
        const inlineComTokens = [];
        const stopIndex = this.getStopTokenIndexOf(ctx);
        for (let i = ctx.start.tokenIndex; i < stopIndex; ++i) {
            const type = tokenStream.tokens[i].type;
            if (type !== COMMA && type !== SEMICOLON && type !== ML_COM && type !== SL_COM && type !== NEWLINE) {
                let rightTokens = getHiddenTokens({ tokenStream, tokenIndex: i, direction: "right", NEWLINE });
                if (rightTokens) {
                    rightTokens = rightTokens.filter(token => (token.type === ML_COM || token.type === SL_COM) && !this.consumed.has(token));
                    this.consumeTokens(rightTokens);
                    inlineComTokens.push(...rightTokens);
                }
            }
        }
        const lastTokens = getHiddenTokens({ tokenStream, tokenIndex: stopIndex, direction: "right", NEWLINE });
        if (lastTokens) {
            const lastToken = lastTokens.find(token => ((token.type === ML_COM || token.type === SL_COM) && !this.consumed.has(token))
                || token.type === NEWLINE);
            if (lastToken && (lastToken.type === ML_COM || lastToken.type === SL_COM)) {
                this.consumeTokens([lastToken]);
                inlineComTokens.push(lastToken);
            }
        }
        return this.inlineTokensToGrabbedComments(inlineComTokens);
    }
    getStopTokenIndexOf(ruleContext) {
        const { COMMA, SEMICOLON } = this.parsingContext.tokenTypes;
        const tokenAfter = this.parsingContext.tokenStream.tokens[ruleContext.stop.tokenIndex + 1];
        let stopIndex = ruleContext.stop.tokenIndex;
        if (tokenAfter && (tokenAfter.type === COMMA || tokenAfter.type === SEMICOLON))
            ++stopIndex;
        return stopIndex;
    }
    inlineTokensToGrabbedComments(tokens) {
        const { tokenTypes: { SINGLE_LINE_COMMENT: SL_COM } } = this.parsingContext;
        const { source } = this.parsingContext;
        const result = [];
        for (const { start, stop, type } of tokens) {
            let text;
            if (type === SL_COM)
                text = formatInlineComment(source, start, stop);
            else
                text = formatMultiLineComment(source, start, stop, "asInline").text;
            if (text.length > 0)
                result.push({ syntax: type === SL_COM ? "inline" : "classic", text });
        }
        return result;
    }
    consumeTokens(tokens) {
        tokens.forEach(tok => this.consumed.add(tok));
    }
}
exports.default = CommentGrabber;
function keepWholeLineComments(tokens, parsingContext) {
    const { tokenTypes: { SINGLE_LINE_COMMENT: SL_COM, NEWLINE } } = parsingContext;
    let newLineCount = 0;
    return tokens.filter(token => {
        if (token.type === NEWLINE) {
            ++newLineCount;
            return newLineCount === 2;
        }
        const keep = token.type !== SL_COM || newLineCount > 0 || token.tokenIndex === 0;
        newLineCount = 0;
        return keep;
    });
}
function formatInlineComment(source, start, stop, mode) {
    const index = start + (source[start + 2] === " " ? 3 : 2);
    if (mode === "asBlockLine") {
        return source.substring(index, stop + 1).trimRight();
    }
    else
        return source.substring(index, stop + 1).trim();
}
function formatMultiLineComment(source, start, stop, mode) {
    const length = stop - start - 3;
    if (length <= 0)
        return { doc: false, text: "" };
    const doc = mode !== "asInline" && source[start + 2] === "*";
    const raw = source.substr(doc ? start + 3 : start + 2, doc ? length - 1 : length);
    let lines = raw.split("\n");
    if (lines.length === 1)
        lines[0] = lines[0].trim();
    else if (mode !== "asInline" && haveMultiLineCommentPrefix(lines)) {
        lines = lines.map(line => {
            const trimed = line.trim();
            return trimed === "" ? "" : trimed.slice(trimed[1] === " " ? 2 : 1);
        });
        if (lines[0] === "")
            lines.shift();
        if (lines[lines.length - 1] === "")
            lines.pop();
    }
    else {
        if (mode === "asInline")
            lines = lines.map(line => line.trim());
        else
            lines = lines.map(line => line.trimRight());
    }
    const text = lines.join(mode === "asInline" ? "; " : "\n");
    return { doc, text };
}
function haveMultiLineCommentPrefix(lines) {
    const noPrefix = lines.find(line => {
        const trimed = line.trim();
        if (trimed === "")
            return false;
        return trimed[0] !== "*";
    });
    return !noPrefix;
}
/**
 * Similar to the ANTLR API `BufferedTokenStream.prototype.getHiddenTokensToRight`, but with not-hidden NEWLINE
 */
function getHiddenTokens(options) {
    const { tokenStream: { tokens }, tokenIndex, direction, NEWLINE, startIndex, stopIndex } = options;
    const hiddenChannel = 1;
    const len = tokens.length;
    if (tokenIndex < -1 || tokenIndex >= len)
        throw new Error(`Invalid tokenIndex: ${tokenIndex}`);
    const result = [];
    const step = direction === "left" ? -1 : 1;
    for (let i = tokenIndex + step; i >= 0 && i < len; i += step) {
        const token = tokens[i];
        if ((startIndex !== undefined && token.tokenIndex < startIndex)
            || (stopIndex !== undefined && token.tokenIndex > stopIndex)
            || (token.channel !== hiddenChannel && token.type !== NEWLINE))
            break;
        result.push(token);
    }
    if (result.length === 0)
        return null;
    if (direction === "left")
        result.reverse();
    return result;
}
