"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
class AstImportTool {
    constructor(path, moduleLoader) {
        this.path = path;
        this.moduleLoader = moduleLoader;
        this.fromModules = new Map();
        this.importedIdentifiers = new Set();
        this.importedNamespaces = new Set();
    }
    addImport(astNode) {
        const ifm = this.getImportedFromModule(astNode.from);
        if (astNode.whichImport === "namespaced")
            this.addNamespacedImport(astNode.asNamespace, ifm);
        else {
            if (astNode.namedMembers) {
                for (const member of astNode.namedMembers)
                    this.addNamedImport(Object.assign({}, member), ifm);
            }
        }
    }
    addInlineImport(astNode) {
        const ifm = this.getImportedFromModule(astNode.from);
        ifm.inlineMembers.add(astNode.exportedName);
    }
    async load() {
        for (const [from, ifm] of this.fromModules.entries()) {
            const factory = await this.moduleLoader({ from, relativeToModule: this.path });
            this.checkExportedNames(ifm, factory);
        }
    }
    createRtoImports() {
        const imports = [];
        const namespacedImports = [];
        for (const [from, ifm] of this.fromModules.entries()) {
            const { namedMembers, namespaces } = ifm;
            if (namedMembers.length > 0) {
                imports.push({
                    from,
                    namedMembers
                });
            }
            namespaces.forEach(asNamespace => namespacedImports.push({ from, asNamespace }));
        }
        const result = {};
        if (imports.length > 0)
            result.imports = imports;
        if (namespacedImports.length > 0)
            result.namespacedImports = namespacedImports;
        return result;
    }
    findImportedMember(fullName) {
        const dotIndex = fullName.indexOf(".");
        if (dotIndex !== -1) {
            const refName = fullName.slice(0, dotIndex);
            const namespace = fullName.slice(dotIndex + 1);
            if (!this.importedNamespaces.has(namespace))
                throw new Error(`Unknown namespace: ${namespace}`);
            return { refName, namespace };
        }
        return this.importedIdentifiers.has(fullName) ? { refName: fullName } : undefined;
    }
    inlineImport({ exportedName }) {
        return { refName: exportedName };
    }
    addNamespacedImport(asNamespace, ifm) {
        if (this.importedNamespaces.has(asNamespace) || this.importedIdentifiers.has(asNamespace))
            throw new Error(`Duplicate identifier '${asNamespace}'`);
        this.importedNamespaces.add(asNamespace);
        ifm.namespaces.push(asNamespace);
    }
    addNamedImport(member, ifm) {
        const name = member.as || member.name;
        if (this.importedNamespaces.has(name) || this.importedIdentifiers.has(name))
            throw new Error(`Duplicate identifier '${name}'`);
        this.importedIdentifiers.add(name);
        ifm.namedMembers.push(Object.assign({}, member));
    }
    getImportedFromModule(from) {
        let ifm = this.fromModules.get(from);
        if (!ifm) {
            ifm = {
                namedMembers: [],
                namespaces: [],
                inlineMembers: new Set(),
            };
            this.fromModules.set(from, ifm);
        }
        return ifm;
    }
    checkExportedNames(ifm, factory) {
        for (const { name } of ifm.namedMembers) {
            if (!factory.hasExportedNamedType(name))
                throw new Error(`Module '${factory.getModulePath()}' has no exported member '${name}'.`);
        }
        for (const name of ifm.inlineMembers) {
            if (!factory.hasExportedNamedType(name))
                throw new Error(`Module '${factory.getModulePath()}' has no exported member '${name}'.`);
        }
    }
}
exports.default = AstImportTool;
