"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
class InlineImportScanner {
    constructor(importTool) {
        this.importTool = importTool;
        this.typeScans = {
            inlineImport: astNode => this.registerInlineImport(astNode),
            literal: () => { },
            array: astNode => this.scanArrayType(astNode),
            tuple: astNode => this.scanTupleType(astNode),
            keyof: astNode => this.scanKeyofType(astNode),
            member: astNode => this.scanMemberType(astNode),
            composite: astNode => this.scanCompositeType(astNode),
            genericInstance: astNode => this.scanGenericInstance(astNode),
            function: astNode => this.scanFunctionType(astNode),
            interface: astNode => this.scanInterface(astNode),
        };
    }
    scan(astNode) {
        if (typeof astNode === "string")
            return;
        const scan = this.typeScans[astNode.whichType];
        if (!scan)
            throw new Error(`Unexpected whichType to scan: ${astNode.whichType}`);
        return scan(astNode);
    }
    registerInlineImport(astNode) {
        this.importTool.addInlineImport(astNode);
    }
    scanArrayType(astNode) {
        this.scan(astNode.itemType);
    }
    scanCompositeType(astNode) {
        astNode.types.forEach(child => this.scan(child));
    }
    scanGenericInstance(astNode) {
        astNode.parameterTypes.forEach(child => this.scan(child));
    }
    scanKeyofType(astNode) {
        this.scan(astNode.type);
    }
    scanMemberType(astNode) {
        this.scan(astNode.parentType);
    }
    scanTupleType(astNode) {
        if (astNode.itemTypes)
            astNode.itemTypes.forEach(child => this.scan(child));
    }
    scanFunctionType(astNode) {
        this.scan(astNode.returnType);
        if (astNode.parameters)
            this.scanFunctionParameters(astNode.parameters);
        if (astNode.generic)
            this.scanGenericParameters(astNode.generic);
    }
    scanFunctionParameters(astNodes) {
        return astNodes.forEach(({ type }) => {
            if (type)
                this.scan(type);
        });
    }
    scanGenericParameters(astNodes) {
        return astNodes.forEach(({ extendsType, defaultType }) => {
            if (extendsType)
                this.scan(extendsType);
            if (defaultType)
                this.scan(defaultType);
        });
    }
    scanInterface(astNode) {
        if (!astNode.entries)
            return;
        for (const entry of astNode.entries) {
            if (entry.whichEntry === "indexSignature")
                this.scanIndexSignature(entry);
            else if (entry.whichEntry === "mappedIndexSignature")
                this.scanMappedIndexSignature(entry);
            else if (entry.whichEntry === "property")
                this.scanProperty(entry);
            else if (entry.whichEntry === "functionProperty")
                this.scanPropertyFromFunctionProperty(entry);
        }
    }
    scanProperty(entry) {
        this.scan(entry.type);
    }
    scanPropertyFromFunctionProperty(entry) {
        if (entry.parameters)
            this.scanFunctionParameters(entry.parameters);
        if (entry.returnType)
            this.scan(entry.returnType);
        if (entry.generic)
            this.scanGenericParameters(entry.generic);
    }
    scanIndexSignature(entry) {
        this.scan(entry.type);
    }
    scanMappedIndexSignature(entry) {
        this.scan(entry.keyInType);
        this.scan(entry.type);
    }
}
exports.default = InlineImportScanner;
