"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const AstImportTool_1 = require("./AstImportTool");
const InlineImportScanner_1 = require("./InlineImportScanner");
const tsTypeNames = new Set(["any", "unknown", "object", "void", "never"]);
const primitiveTypeNames = new Set(["string", "number", "bigint", "boolean", "symbol", "undefined", "null"]);
const defaultGlobals = new Set(["String", "Number", "Bigint", "Boolean", "Symbol", "Date"]);
class RtoModuleFactory {
    constructor(ast, modulePath, options = {}) {
        this.ast = ast;
        this.modulePath = modulePath;
        this.rtoTypeCreators = {
            array: astNode => this.createRtoArrayType(astNode),
            literal: astNode => this.createRtoLiteralType(astNode),
            composite: astNode => this.createRtoCompositeType(astNode),
            genericInstance: astNode => this.createRtoGenericInstance(astNode),
            keyof: astNode => this.createRtoKeyofType(astNode),
            member: astNode => this.createRtoMemberType(astNode),
            tuple: astNode => this.createRtoTupleType(astNode),
            function: astNode => this.createRtoFunctionType(astNode),
            inlineImport: astNode => this.createRtoImportedRefFromInline(astNode),
            interface: astNode => this.createRtoInterface(astNode),
        };
        this.namedTypeList = [];
        this.namedTypes = new Map();
        this.globals = new Set(options.defineGlobals ? options.defineGlobals(defaultGlobals) : defaultGlobals);
        if (ast.declarations)
            ast.declarations.forEach(astDecl => this.registerAstDeclaration(astDecl));
    }
    hasExportedNamedType(name) {
        const namedType = this.namedTypes.get(name);
        return !!(namedType && namedType.exported);
    }
    getModulePath() {
        if (!this.modulePath)
            throw new Error(`Missing module path`);
        return this.modulePath;
    }
    async loadImports(moduleLoader) {
        this.importTool = new AstImportTool_1.default(this.getModulePath(), moduleLoader);
        if (this.ast.declarations) {
            const inlineScanner = new InlineImportScanner_1.default(this.importTool);
            for (const astDecl of this.ast.declarations) {
                if (astDecl.whichDeclaration === "import")
                    this.importTool.addImport(astDecl);
                else if (astDecl.whichDeclaration === "interface")
                    inlineScanner.scan(astDecl);
                else if (astDecl.whichDeclaration === "type")
                    inlineScanner.scan(astDecl.type);
                else if (astDecl.whichDeclaration !== "comment")
                    throw new Error(`Invalid whichDeclaration: ${astDecl.whichDeclaration}`);
            }
        }
        await this.importTool.load();
    }
    createRtoModule() {
        if (this.ast.declarations) {
            this.ast.declarations.forEach(astDecl => {
                if (astDecl.whichDeclaration === "interface" || astDecl.whichDeclaration === "type")
                    this.fillRtoNamed(astDecl);
                else if (astDecl.whichDeclaration === "import" && !this.importTool)
                    throw new Error(`Imports are not loaded`);
            });
        }
        const module = {};
        if (this.importTool) {
            Object.assign(module, this.importTool.createRtoImports());
        }
        if (this.namedTypeList.length > 0)
            module.namedTypes = this.namedTypeList;
        return module;
    }
    registerAstDeclaration(astNode) {
        switch (astNode.whichDeclaration) {
            case "interface":
            case "type":
                const namedType = this.createRtoBaseNamedType(astNode);
                this.namedTypes.set(namedType.name, namedType);
                this.namedTypeList.push(namedType);
                break;
            case "import":
            case "comment":
                break;
            default:
                throw new Error(`Invalid whichDeclaration: ${astNode.whichDeclaration}`);
        }
    }
    createRtoBaseNamedType(astNode) {
        const result = {
            name: astNode.name
        };
        if (astNode.exported)
            result.exported = astNode.exported;
        if (astNode.generic)
            result.generic = this.createRtoGenericParameters(astNode.generic);
        if (astNode.docComment)
            result.docComment = astNode.docComment;
        return result;
    }
    fillRtoNamed(astNode) {
        const base = this.getBaseNamedType(astNode.name);
        const type = astNode.whichDeclaration === "interface"
            ? (astNode.extends ? this.convertExtendsToRtoIntersection(astNode) : this.createRtoInterface(astNode))
            : this.createRtoType(astNode.type);
        Object.assign(base, type);
    }
    convertExtendsToRtoIntersection(astNode) {
        const typeNames = astNode.extends;
        const types = typeNames.map(child => this.createRtoType(child));
        types.push(this.createRtoInterface(astNode));
        return {
            kind: "composite",
            op: "intersection",
            types
        };
    }
    createRtoType(astNode) {
        if (typeof astNode === "string") {
            const kindOfName = this.findKindOfName(astNode);
            if (kindOfName)
                return createRtoTypeName(kindOfName, astNode);
            if (this.namedTypes.get(astNode))
                return createRtoLocalTypeRef(astNode);
            if (this.importTool) {
                const importRef = this.importTool.findImportedMember(astNode);
                if (importRef)
                    return createRtoImportedTypeRef(importRef);
            }
            throw new Error(`Unexpected type: ${astNode}`);
        }
        else {
            const creator = this.rtoTypeCreators[astNode.whichType];
            if (!creator)
                throw new Error(`Unexpected whichType: ${astNode.whichType}`);
            return creator(astNode);
        }
    }
    createRtoArrayType(astNode) {
        return {
            kind: "array",
            itemType: this.createRtoType(astNode.itemType)
        };
    }
    createRtoLiteralType(astNode) {
        return {
            kind: "literal",
            literal: astNode.literal
        };
    }
    createRtoCompositeType(astNode) {
        return {
            kind: "composite",
            op: astNode.op,
            types: astNode.types.map(child => this.createRtoType(child))
        };
    }
    createRtoGenericInstance(astNode) {
        return {
            kind: "genericInstance",
            genericName: astNode.genericName,
            parameterTypes: astNode.parameterTypes.map(child => this.createRtoType(child))
        };
    }
    createRtoKeyofType(astNode) {
        return {
            kind: "keyof",
            type: this.createRtoType(astNode.type)
        };
    }
    createRtoMemberType(astNode) {
        return {
            kind: "member",
            parentType: this.createRtoType(astNode.parentType),
            memberName: typeof astNode.memberName === "string" ? astNode.memberName : {
                literal: astNode.memberName.literal
            }
        };
    }
    createRtoTupleType(astNode) {
        const type = {
            kind: "tuple",
        };
        if (astNode.itemTypes)
            type.itemTypes = astNode.itemTypes.map(child => this.createRtoType(child));
        return type;
    }
    createRtoFunctionType(astNode) {
        const type = {
            kind: "function",
            returnType: this.createRtoType(astNode.returnType),
        };
        if (astNode.parameters)
            type.parameters = this.createRtoFunctionParameters(astNode.parameters);
        if (astNode.generic)
            type.generic = this.createRtoGenericParameters(astNode.generic);
        return type;
    }
    createRtoFunctionParameters(astNodes) {
        return astNodes.map(({ name, type }) => {
            const param = { name };
            if (type)
                param.type = this.createRtoType(type);
            return param;
        });
    }
    createRtoGenericParameters(astNodes) {
        return astNodes.map(({ name, extendsType, defaultType }) => {
            const param = { name };
            if (extendsType)
                param.extendsType = this.createRtoType(extendsType);
            if (defaultType)
                param.defaultType = this.createRtoType(defaultType);
            return param;
        });
    }
    createRtoImportedRefFromInline(astNode) {
        if (!this.importTool)
            throw new Error(`Imports are not loaded`);
        return createRtoImportedTypeRef(this.importTool.inlineImport(astNode));
    }
    createRtoInterface(astNode) {
        const result = {
            kind: "interface"
        };
        if (astNode.entries) {
            for (const entry of astNode.entries) {
                if (entry.whichEntry === "indexSignature") {
                    result.indexSignature = this.createRtoIndexSignature(entry);
                }
                else if (entry.whichEntry === "mappedIndexSignature") {
                    result.mappedIndexSignature = this.createRtoMappedIndexSignature(entry);
                }
                else if (entry.whichEntry === "property") {
                    if (!result.properties)
                        result.properties = [];
                    result.properties.push(this.createRtoProperty(entry));
                }
                else if (entry.whichEntry === "functionProperty") {
                    if (!result.properties)
                        result.properties = [];
                    result.properties.push(this.createRtoPropertyFromFunctionProperty(entry));
                }
            }
        }
        return result;
    }
    createRtoProperty(entry) {
        const property = {
            name: entry.name,
            type: this.createRtoType(entry.type),
        };
        if (entry.optional)
            property.optional = entry.optional;
        if (entry.readonly)
            property.readonly = entry.readonly;
        if (entry.docComment)
            property.docComment = entry.docComment;
        return property;
    }
    createRtoPropertyFromFunctionProperty(entry) {
        const type = {
            kind: "function",
            returnType: this.createRtoType(entry.returnType || "any"),
        };
        if (entry.parameters)
            type.parameters = this.createRtoFunctionParameters(entry.parameters);
        if (entry.generic)
            type.generic = this.createRtoGenericParameters(entry.generic);
        const property = {
            name: entry.name,
            type,
        };
        if (entry.optional)
            property.optional = entry.optional;
        if (entry.readonly)
            property.readonly = entry.readonly;
        if (entry.docComment)
            property.docComment = entry.docComment;
        return property;
    }
    createRtoIndexSignature(entry) {
        const indexSignature = {
            keyName: entry.keyName,
            keyType: entry.keyType,
            type: this.createRtoType(entry.type)
        };
        if (entry.optional)
            indexSignature.optional = entry.optional;
        if (entry.readonly)
            indexSignature.readonly = entry.readonly;
        if (entry.docComment)
            indexSignature.docComment = entry.docComment;
        return indexSignature;
    }
    createRtoMappedIndexSignature(entry) {
        const mappedIndexSignature = {
            keyName: entry.keyName,
            keyInType: this.createRtoType(entry.keyInType),
            type: this.createRtoType(entry.type)
        };
        if (entry.optional)
            mappedIndexSignature.optional = entry.optional;
        if (entry.readonly)
            mappedIndexSignature.readonly = entry.readonly;
        if (entry.docComment)
            mappedIndexSignature.docComment = entry.docComment;
        return mappedIndexSignature;
    }
    getBaseNamedType(name) {
        const result = this.namedTypes.get(name);
        if (!result)
            throw new Error(`Unknown named type: ${name}`);
        return result;
    }
    findKindOfName(typeName) {
        if (tsTypeNames.has(typeName))
            return "ts";
        if (primitiveTypeNames.has(typeName))
            return "primitive";
        if (this.globals.has(typeName))
            return "global";
    }
}
exports.default = RtoModuleFactory;
function createRtoTypeName(group, refName) {
    return {
        kind: "name",
        group,
        refName
    };
}
function createRtoLocalTypeRef(refName) {
    return {
        kind: "localRef",
        refName
    };
}
function createRtoImportedTypeRef(ref) {
    return Object.assign({ kind: "importedRef" }, ref);
}
