// -- API definitions --

export interface ParseTypeOnlyOptions {
  source: string
  return?: {
    /**
     * Default value is `false`.
     */
    freeze?: boolean
  }
}
export declare function parseTypeOnly(options: ParseTypeOnlyOptions): TypeOnlyAst
export interface GenerateRtoModulesOptions {
  modulePaths: string[]
  readFiles?: {
    sourceDir: string
    encoding?: string
  }
  astProvider?: TypeOnlyAstProvider
  defineGlobals?: (globals: Set<string>) => Set<string>
  writeFiles?: boolean | {
    encoding?: string
    outputDir?: string
    /**
     * The indentation parameter of `JSON.stringify`.
     */
    prettify?: number | string
  }
  returnRtoModules?: boolean | {
    /**
     * Default value is `false`.
     */
    freeze?: boolean
  }
}
export declare type TypeOnlyAstProvider = (modulePath: string) => Promise<TypeOnlyAst> | TypeOnlyAst
export declare function generateRtoModules(options: GenerateRtoModulesOptions): Promise<RtoModules | undefined>
export interface CreateStandaloneRtoModuleOptions {
  ast: TypeOnlyAst
  defineGlobals?: (globals: Set<string>) => Set<string>
  /**
   * Default value is `false`.
   */
  freeze?: boolean
}
export declare function createStandaloneRtoModule(options: CreateStandaloneRtoModuleOptions): RtoModule

// -- RTO definitions --

export interface RtoModules {
  [modulePath: string]: RtoModule
}

/**
 * Raw TypeOnly Module
 */
export interface RtoModule {
  imports?: RtoImport[]
  namespacedImports?: RtoNamespacedImport[]
  namedTypes?: RtoNamedType[]
}

export interface RtoImport {
  from: string
  namedMembers?: RtoImportNamedMember[]
}

export interface RtoImportNamedMember {
  name: string
  as?: string
}

export interface RtoNamespacedImport {
  from: string
  asNamespace: string
}

export type RtoNamedType = RtoType & RtoBaseNamedType

export type RtoNamedTypeName = RtoTypeName & RtoBaseNamedType
export type RtoNamedGenericParameterName = RtoGenericParameterName & RtoBaseNamedType
export type RtoNamedLocalTypeRef = RtoLocalTypeRef & RtoBaseNamedType
export type RtoNamedImportedTypeRef = RtoImportedTypeRef & RtoBaseNamedType
export type RtoNamedLiteralType = RtoLiteralType & RtoBaseNamedType
export type RtoNamedCompositeType = RtoCompositeType & RtoBaseNamedType
export type RtoNamedTupleType = RtoTupleType & RtoBaseNamedType
export type RtoNamedArrayType = RtoArrayType & RtoBaseNamedType
export type RtoNamedGenericInstance = RtoGenericInstance & RtoBaseNamedType
export type RtoNamedFunctionType = RtoFunctionType & RtoBaseNamedType
export type RtoNamedKeyofType = RtoKeyofType & RtoBaseNamedType
export type RtoNamedMemberType = RtoMemberType & RtoBaseNamedType
export type RtoNamedInterface = RtoInterface & RtoBaseNamedType

export type RtoType = RtoTypeName
  | RtoGenericParameterName
  | RtoLocalTypeRef
  | RtoImportedTypeRef
  | RtoLiteralType
  | RtoCompositeType
  | RtoTupleType
  | RtoArrayType
  | RtoGenericInstance
  | RtoFunctionType
  | RtoKeyofType
  | RtoMemberType
  | RtoInterface

export interface RtoBaseNamedType extends RtoCommentable {
  exported?: boolean
  name: string
  generic?: RtoGenericParameter[]
}

export interface RtoGenericParameter {
  name: string
  extendsType?: RtoType
  defaultType?: RtoType
}

export interface RtoTypeName {
  kind: "name"
  group: "ts" | "primitive" | "global"
  refName: RtoSpecialTypeName | RtoPrimitiveTypeName | string
}

export type RtoSpecialTypeName = "any" | "unknown" | "object" | "void" | "never"
export type RtoPrimitiveTypeName = "string" | "number" | "bigint" | "boolean" | "symbol" | "undefined" | "null"

export interface RtoGenericParameterName {
  kind: "genericParameterName"
  genericParameterName: string
}

export interface RtoLocalTypeRef {
  kind: "localRef"
  refName: string
}

export interface RtoImportedTypeRef {
  kind: "importedRef"
  refName: string
  namespace?: string
}

export interface RtoLiteralType {
  kind: "literal"
  literal: string | number | bigint | boolean
}

export interface RtoCompositeType {
  kind: "composite"
  op: "union" | "intersection"
  types: RtoType[]
}

export interface RtoTupleType {
  kind: "tuple"
  itemTypes?: RtoType[]
}

export interface RtoArrayType {
  kind: "array"
  itemType: RtoType
}

export interface RtoGenericInstance {
  kind: "genericInstance"
  genericName: string
  parameterTypes: RtoType[]
}

export interface RtoFunctionType {
  kind: "function"
  parameters?: RtoFunctionParameter[]
  returnType: RtoType
  generic?: RtoGenericParameter[]
}

export interface RtoFunctionParameter {
  name: string
  type?: RtoType
}

export interface RtoKeyofType {
  kind: "keyof"
  type: RtoType
}

export interface RtoMemberType {
  kind: "member"
  parentType: RtoType
  memberName: string | RtoMemberNameLiteral
}

export interface RtoMemberNameLiteral {
  literal: string | number
}

export interface RtoInterface {
  kind: "interface"
  indexSignature?: RtoIndexSignature
  mappedIndexSignature?: RtoMappedIndexSignature
  properties?: RtoProperty[]
}

export interface RtoProperty extends RtoCommentable {
  name: string
  type: RtoType
  optional?: boolean
  readonly?: boolean
}

export interface RtoIndexSignature extends RtoCommentable {
  keyName: string
  keyType: "string" | "number"
  type: RtoType
  optional?: boolean
  readonly?: boolean
}

export interface RtoMappedIndexSignature extends RtoCommentable {
  keyName: string
  keyInType: RtoType
  type: RtoType
  optional?: boolean
  readonly?: boolean
}

export interface RtoCommentable {
  /**
   * A multiline string.
   */
  docComment?: string
}

// -- AST definitions --

export interface TypeOnlyAst {
  declarations?: AstDeclaration[]
}

export type AstDeclaration = AstImport
  | AstNamedInterface
  | AstNamedType
  | AstStandaloneComment

export type AstImport = AstClassicImport | AstNamespacedImport

export interface AstClassicImport extends AstCommentable {
  whichDeclaration: "import"
  whichImport: "classic"
  from: string
  namedMembers?: AstImportNamedMember[]
}

export interface AstImportNamedMember {
  name: string
  as?: string
}

export interface AstNamespacedImport extends AstCommentable {
  whichDeclaration: "import"
  whichImport: "namespaced"
  from: string
  asNamespace: string
}

export interface AstInterface {
  whichType: "interface"
  entries?: AstInterfaceEntry[]
}

export interface AstNamedInterface extends AstInterface, AstCommentable {
  whichDeclaration: "interface"
  name: string
  exported?: boolean
  extends?: string[]
  generic?: AstGenericParameter[]
}

export type AstInterfaceEntry = AstProperty
  | AstFunctionProperty
  | AstIndexSignature
  | AstMappedIndexSignature
  | AstStandaloneInterfaceComment

export interface AstProperty extends AstCommentable {
  whichEntry: "property"
  optional?: boolean
  readonly?: boolean
  name: string
  type: AstType
}

export interface AstFunctionProperty extends AstCommentable {
  whichEntry: "functionProperty"
  optional?: boolean
  readonly?: boolean
  name: string
  parameters?: AstFunctionParameter[]
  returnType?: AstType
  generic?: AstGenericParameter[]
}

export interface AstIndexSignature extends AstCommentable {
  whichEntry: "indexSignature"
  keyName: string
  keyType: "string" | "number"
  type: AstType
  optional?: boolean
  readonly?: boolean
}

export interface AstMappedIndexSignature extends AstCommentable {
  whichEntry: "mappedIndexSignature"
  keyName: string
  keyInType: AstType
  type: AstType
  optional?: boolean
  readonly?: boolean
}

export interface AstNamedType extends AstCommentable {
  whichDeclaration: "type"
  name: string
  type: AstType
  exported?: boolean
  generic?: AstGenericParameter[]
}

export type AstType = string
  | AstLiteralType
  | AstInterface
  | AstCompositeType
  | AstTupleType
  | AstArrayType
  | AstGenericInstance
  | AstFunctionType
  | AstKeyofType
  | AstMemberType
  | AstInlineImportType

export interface AstLiteralType {
  whichType: "literal"
  literal: string | number | boolean | bigint
  stringDelim?: "\"" | "'" | "`"
}

export interface AstCompositeType {
  whichType: "composite"
  op: "union" | "intersection"
  types: AstType[]
}

export interface AstTupleType {
  whichType: "tuple"
  itemTypes?: AstType[]
}

export interface AstArrayType {
  whichType: "array"
  itemType: AstType
  genericSyntax?: boolean
}

export interface AstGenericInstance {
  whichType: "genericInstance"
  genericName: string
  parameterTypes: AstType[]
}

export interface AstFunctionType {
  whichType: "function"
  parameters?: AstFunctionParameter[]
  returnType: AstType
  generic?: AstGenericParameter[]
}

export interface AstFunctionParameter {
  name: string
  type?: AstType
}

export interface AstKeyofType {
  whichType: "keyof"
  type: AstType
}

export interface AstMemberType {
  whichType: "member"
  parentType: AstType
  memberName: string | AstMemberNameLiteral
}

export interface AstMemberNameLiteral {
  literal: string | number
  stringDelim?: "\"" | "'" | "`"
}

export interface AstInlineImportType {
  whichType: "inlineImport"
  from: string
  exportedName: string
}

export interface AstGenericParameter {
  name: string
  extendsType?: AstType
  defaultType?: AstType
}

export interface AstStandaloneComment {
  whichDeclaration: "comment"
  /**
   * A multiline string.
   */
  text: string
  syntax: "inline" | "classic"
}

export interface AstStandaloneInterfaceComment {
  whichEntry: "comment"
  /**
   * A multiline string.
   */
  text: string
  syntax: "inline" | "classic"
}

export interface AstCommentable {
  /**
   * A multiline string.
   */
  docComment?: string
  inlineComments?: AstInlineComment[]
}

export interface AstInlineComment {
  /**
   * A single line string.
   */
  text: string
  syntax: "inline" | "classic"
}